#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <math.h>

#define POOL_SIZE 2
#define BLOCK_SIZE 256

// Kernel using a 3D grid mapping:
//   gridDim.x -> batch index
//   gridDim.y -> output channel index
//   gridDim.z -> blocks of pooling spatial indices

__global__ void conv_pool_sigmoid_sum_kernel(
    const float* __restrict__ input,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* output,
    const int batch_size,
    const int in_channels,
    const int out_channels,
    const int height,
    const int width,
    const int kernel_size,
    const int pool_out_height,
    const int pool_out_width
) {
    // Mapping: batch index and output channel from grid.x and grid.y
    int batch = blockIdx.x;
    int oc = blockIdx.y;
    
    // Total number of pooling output positions
    int pool_total = pool_out_height * pool_out_width;
    
    // Each block in grid.z covers a segment of pooling indices
    int idx = blockIdx.z * blockDim.x + threadIdx.x;
    
    float value = 0.0f;
    if (idx < pool_total) {
        // Calculate pooling spatial coordinates
        int pool_h = idx / pool_out_width;
        int pool_w = idx % pool_out_width;

        float conv_result = bias[oc];

        // Perform convolution over input channels and kernel window
        #pragma unroll
        for (int ic = 0; ic < in_channels; ic++) {
            // Compute base offsets for input and weight
            int input_channel_offset = ((batch * in_channels) + ic) * height * width;
            int weight_channel_offset = ((oc * in_channels) + ic) * kernel_size * kernel_size;
            
            int base_h = pool_h * POOL_SIZE;
            int base_w = pool_w * POOL_SIZE;

            #pragma unroll
            for (int kh = 0; kh < kernel_size; kh++) {
                int h_in = base_h + kh;
                #pragma unroll
                for (int kw = 0; kw < kernel_size; kw++) {
                    int w_in = base_w + kw;
                    if (h_in < height && w_in < width) {
                        int input_index = input_channel_offset + h_in * width + w_in;
                        int weight_index = weight_channel_offset + kh * kernel_size + kw;
                        conv_result += input[input_index] * weight[weight_index];
                    }
                }
            }
        }
        
        // Average pooling and apply sigmoid activation
        conv_result *= (1.0f / (POOL_SIZE * POOL_SIZE));
        value = 1.0f / (1.0f + expf(-conv_result));
    } else {
        value = 0.0f;
    }

    // Block-level reduction using shared memory
    extern __shared__ float sdata[];
    sdata[threadIdx.x] = value;
    __syncthreads();

    // Standard reduction in shared memory
    for (unsigned int s = blockDim.x / 2; s > 0; s >>= 1) {
        if (threadIdx.x < s) {
            sdata[threadIdx.x] += sdata[threadIdx.x + s];
        }
        __syncthreads();
    }

    // Thread 0 of each block atomically adds the partial sum to global output for the batch
    if (threadIdx.x == 0) {
        atomicAdd(&output[batch], sdata[0]);
    }
}

// Host function to setup grid and launch kernel
torch::Tensor forward(
    torch::Tensor input,
    torch::Tensor weight,
    torch::Tensor bias
) {
    const int batch_size = input.size(0);
    const int in_channels = input.size(1);
    const int height = input.size(2);
    const int width = input.size(3);
    const int out_channels = weight.size(0);
    const int kernel_size = weight.size(2);

    // Compute convolution output dimensions
    int out_height = height - kernel_size + 1;
    int out_width = width - kernel_size + 1;
    
    // Compute pooling output dimensions
    int pool_out_height = out_height / POOL_SIZE;
    int pool_out_width = out_width / POOL_SIZE;
    int pool_total = pool_out_height * pool_out_width;

    // Determine grid dimensions using 3D indexing
    // grid.x: batch dimension
    // grid.y: out_channels dimension
    // grid.z: covers pooled spatial positions, chunked by BLOCK_SIZE
    int blocks_z = (pool_total + BLOCK_SIZE - 1) / BLOCK_SIZE;
    dim3 blocks(batch_size, out_channels, blocks_z);
    int threads = BLOCK_SIZE;

    auto output = torch::zeros({batch_size}, input.options());

    // Launch kernel with shared memory per block
    conv_pool_sigmoid_sum_kernel<<<blocks, threads, BLOCK_SIZE * sizeof(float)>>>(
        input.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        output.data_ptr<float>(),
        batch_size,
        in_channels,
        out_channels,
        height,
        width,
        kernel_size,
        pool_out_height,
        pool_out_width
    );

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "3D indexing optimized convolution + pool + sigmoid + sum forward");
}
